/**h* DesfireAPI/Management
 *
 * NAME
 *   DesfireAPI :: Card management functions
 *
 * COPYRIGHT
 *   (c) 2009 SpringCard - www.springcard.com
 *
 * DESCRIPTION
 *   Implementation of management functions to personalize or format the
 *   DESFIRE card.
 *
 **/
#include "sprox_desfire_i.h"

/**f* DesfireAPI/GetFreeMemory
 *
 * NAME
 *   GetFreeMemory
 *
 * DESCRIPTION
 *   Reads out the number of available bytes on the PICC
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_GetFreeMemory(DWORD *pdwFreeBytes);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_GetFreeMemory(SPROX_INSTANCE rInst,
 *                                      DWORD *pdwFreeBytes);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_GetFreeMemory(SCARDHANDLE hCard,
 *                                    DWORD *pdwFreeBytes);
 *
 * INPUTS
 *   DWORD *pdwFreeBytes : number of free bytes on the PICC
 *
 * RETURNS
 *   DF_OPERATION_OK : operation succeeded
 *   Other code if internal or communication error has occured. 
 *
 * NOTES
 *   This command can be issued without valid authentication.
 *
 **/
SPROX_API_FUNC(Desfire_GetFreeMemory) (SPROX_PARAM  DWORD *pdwFreeBytes)
{
  SPROX_RC   status;
  SPROX_DESFIRE_GET_CTX();

  /* Create the info block containing the command code and the key number argument. */
  ctx->xfer_buffer[INF + 0] = DF_GET_FREE_MEMORY;
  ctx->xfer_length = 1;

  /* Communicate the info block to the card and check the operation's return status. */
  /* The returned info block must contain 4 bytes, the status code and the requested */
  /* information.                                                                    */
  status = SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  4, COMPUTE_COMMAND_CMAC | CHECK_RESPONSE_CMAC | WANTS_OPERATION_OK);
  if (status != DF_OPERATION_OK)
    return status;

  /* Get the actual value. */
  if (pdwFreeBytes != NULL)
  {
    *pdwFreeBytes  = 0;
    *pdwFreeBytes += ctx->xfer_buffer[INF + 3]; *pdwFreeBytes <<= 8;
    *pdwFreeBytes += ctx->xfer_buffer[INF + 2]; *pdwFreeBytes <<= 8;
    *pdwFreeBytes += ctx->xfer_buffer[INF + 1];
  }

  /* Success. */
  return DF_OPERATION_OK;
}


/**f* DesfireAPI/FormatPICC
 *
 * NAME
 *   FormatPICC
 *
 * DESCRIPTION
 *   Releases the DesFire card user memory
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_FormatPICC(void);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_FormatPICC(SPROX_INSTANCE rInst);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_FormatPICC(SCARDHANDLE hCard);
 *
 * RETURNS
 *   DF_OPERATION_OK    : format succeeded
 *   Other code if internal or communication error has occured. 
 *
 * NOTES
 *   All applications are deleted and all files within those applications  are deleted.
 *   This command always requires a preceding authentication with the DesFire card master key.
 *
 **/
SPROX_API_FUNC(Desfire_FormatPICC) (SPROX_PARAM_V)
{
  SPROX_DESFIRE_GET_CTX();
  
  /* Create the info block containing the command code. */
  ctx->xfer_buffer[INF + 0] = DF_FORMAT_PICC;
  ctx->xfer_length = 1;

  /* Communicate the info block to the card and check the operation's return status. */
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  0, WANTS_OPERATION_OK);
}

/**f* DesfireAPI/CreateApplication
 *
 * NAME
 *   CreateApplication
 *
 * DESCRIPTION
 *   Create a new application on the DesFire card
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_CreateApplication(DWORD aid,
 *                                         BYTE key_setting_1,
 *                                         BYTE key_setting_2);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_CreateApplication(SPROX_INSTANCE rInst,
 *                                          DWORD aid,
 *                                          BYTE key_setting_1,
 *                                          BYTE key_setting_2);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_CreateApplication(SCARDHANDLE hCard,
 *                                        DWORD aid,
 *                                        BYTE key_setting_1,
 *                                        BYTE key_setting_2);
 *
 * INPUTS
 *   DWORD aid          : Application IDentifier
 *   BYTE key_setting_1 : Settings of the Application master key (see chapter 4.3.2 of datasheet
 *                        of mifare DesFire MF3ICD40 for more information)
 *   BYTE key_setting_2 : Number of keys that can be stored within the application for 
 *                        cryptographic purposes, plus flags to specify cryptographic method and
 *                        to enable giving ISO names to the EF.
 *
 * RETURNS
 *   DF_OPERATION_OK : application created successfully
 *   Other code if internal or communication error has occured. 
 *
 * SEE ALSO
 *   CreateApplicationIso
 *   DeleteApplication
 *   GetApplicationIDs
 *   SelectApplication
 *
 **/
SPROX_API_FUNC(Desfire_CreateApplication) (SPROX_PARAM  DWORD aid, BYTE key_setting_1, BYTE key_setting_2)
{
  SPROX_DESFIRE_GET_CTX();
 
  /* Create the info block containing the command code and the given parameters. */
  ctx->xfer_length = 0;
  ctx->xfer_buffer[ctx->xfer_length++] = DF_CREATE_APPLICATION;
  ctx->xfer_buffer[ctx->xfer_length++] = (BYTE) (aid & 0x000000FF);
  aid >>= 8;
  ctx->xfer_buffer[ctx->xfer_length++] = (BYTE) (aid & 0x000000FF);
  aid >>= 8;
  ctx->xfer_buffer[ctx->xfer_length++] = (BYTE) (aid & 0x000000FF);
  ctx->xfer_buffer[ctx->xfer_length++] = key_setting_1;
  ctx->xfer_buffer[ctx->xfer_length++] = key_setting_2;

  /* Communicate the info block to the card and check the operation's return status. */
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  0, WANTS_OPERATION_OK);
}

/**f* DesfireAPI/CreateApplicationIso
 *
 * NAME
 *   CreateApplicationIso
 *
 * DESCRIPTION
 *   Create a new application on the DesFire card, and defines the ISO identifier
 *   and name of the application
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_CreateApplicationIso(DWORD aid,
 *                                            BYTE key_setting_1,
 *                                            BYTE key_setting_2,
 *                                            WORD iso_df_id,
 *                                            const BYTE iso_df_name[],
 *                                            BYTE iso_df_namelen);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_CreateApplicationIso(SPROX_INSTANCE rInst,
 *                                             DWORD aid,
 *                                             BYTE key_setting_1,
 *                                             BYTE key_setting_2,
 *                                             WORD iso_df_id,
 *                                             const BYTE iso_df_name[],
 *                                             BYTE iso_df_namelen);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_CreateApplicationIso(SCARDHANDLE hCard,
 *                                           DWORD aid,
 *                                           BYTE key_setting_1,
 *                                           BYTE key_setting_2,
 *                                           WORD iso_df_id,
 *                                           const BYTE iso_df_name[],
 *                                           BYTE iso_df_namelen);
 *
 * INPUTS
 *   DWORD aid                : Application IDentifier
 *   BYTE key_setting_1       : Settings of the Application master key (see chapter 4.3.2 of datasheet
 *                              of mifare DesFire MF3ICD40 for more information)
 *   BYTE key_setting_2       : Number of keys that can be stored within the application for 
 *                              cryptographic purposes, plus flags to specify cryptographic method and
 *                              to enable giving ISO names to the EF.
 *   BYTE iso_df_id           : ID of the ISO DF
 *   const BYTE iso_df_name[] : name of the ISO DF
 *   BYTE iso_df_namelen      : length of iso_df_name
 *
 * RETURNS
 *   DF_OPERATION_OK : application created successfully
 *   Other code if internal or communication error has occured. 
 *
 * SEE ALSO
 *   CreateApplication
 *   DeleteApplication
 *   GetApplicationIDs
 *   SelectApplication
 *
 **/
SPROX_API_FUNC(Desfire_CreateApplicationIso) (SPROX_PARAM  DWORD aid, BYTE key_setting_1, BYTE key_setting_2, WORD iso_df_id, const BYTE iso_df_name[16], BYTE iso_df_namelen)
{
  SPROX_DESFIRE_GET_CTX();

  ctx->xfer_length = 0;
  
  /* Create the info block containing the command code and the given parameters. */
  ctx->xfer_buffer[ctx->xfer_length++] = DF_CREATE_APPLICATION;
  ctx->xfer_buffer[ctx->xfer_length++] = (BYTE) (aid & 0x000000FF);
  aid >>= 8;
  ctx->xfer_buffer[ctx->xfer_length++] = (BYTE) (aid & 0x000000FF);
  aid >>= 8;
  ctx->xfer_buffer[ctx->xfer_length++] = (BYTE) (aid & 0x000000FF);
  ctx->xfer_buffer[ctx->xfer_length++] = key_setting_1;
  ctx->xfer_buffer[ctx->xfer_length++] = key_setting_2;
  ctx->xfer_buffer[ctx->xfer_length++] = (BYTE) (iso_df_id); // JDA 
  ctx->xfer_buffer[ctx->xfer_length++] = (BYTE) (iso_df_id >> 8); // JDA

  if (iso_df_name != NULL)
  {
    if (iso_df_namelen == 0)
      iso_df_namelen = strlen(iso_df_name);
    if (iso_df_namelen > 16)
      return DFCARD_LIB_CALL_ERROR;
   
    memcpy(&ctx->xfer_buffer[ctx->xfer_length], iso_df_name, iso_df_namelen);
    ctx->xfer_length += iso_df_namelen;
  }

  /* Communicate the info block to the card and check the operation's return status. */
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  0, WANTS_OPERATION_OK);
}

/**f* DesfireAPI/DeleteApplication
 *
 * NAME
 *   DeleteApplication
 *
 * DESCRIPTION
 *   Permanently deactivates an application on the DesFire card
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_DeleteApplication(DWORD aid);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_DeleteApplication(SPROX_INSTANCE rInst,
 *                                     DWORD aid);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_DeleteApplication(SCARDHANDLE hCard,
 *                                     DWORD aid);
 *
 * INPUTS
 *   DWORD aid                   : Application IDentifier
 *
 * RETURNS
 *   DF_OPERATION_OK    : application deleted successfully
 *   Other code if internal or communication error has occured. 
 *
 * SEE ALSO
 *   CreateApplication
 *   GetApplicationIDs
 *   SelectApplication
 *
 **/
SPROX_API_FUNC(Desfire_DeleteApplication) (SPROX_PARAM  DWORD aid)
{
  SPROX_DESFIRE_GET_CTX();
  
  /* Create the info block containing the command code and the given parameters. */
  ctx->xfer_buffer[INF + 0] = DF_DELETE_APPLICATION;
  ctx->xfer_buffer[INF + 1] = (BYTE) (aid & 0x000000FF);
  aid >>= 8;
  ctx->xfer_buffer[INF + 2] = (BYTE) (aid & 0x000000FF);
  aid >>= 8;
  ctx->xfer_buffer[INF + 3] = (BYTE) (aid & 0x000000FF);
  ctx->xfer_length = 4;

  /* Communicate the info block to the card and check the operation's return status. */
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  0, WANTS_OPERATION_OK);
}

/**f* DesfireAPI/GetApplicationIDs
 *
 * NAME
 *   GetApplicationIDs
 *
 * DESCRIPTION
 *   Returns the Application IDentifiers of all active applications on a DesFire card
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_GetApplicationIDs(BYTE aid_max_count,
 *                                     DWORD aid_list[],
 *                                     BYTE *aid_count);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_GetApplicationIDs(SPROX_INSTANCE rInst,
 *                                     BYTE aid_max_count,
 *                                     DWORD aid_list[],
 *                                     BYTE *aid_count);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_GetApplicationIDs(SCARDHANDLE hCard,
 *                                     BYTE aid_max_count,
 *                                     DWORD aid_list[],
 *                                     BYTE *aid_count);
 *
 * INPUTS
 *   BYTE aid_max_count          : maximum number of Application IDentifiers
 *   DWORD aid_list[]            : Application IDentifier list 
 *   BYTE *aid_count             : number of Application IDentifiers on DesFire card
 *
 * RETURNS
 *   DF_OPERATION_OK    : operation succeeded
 *   Other code if internal or communication error has occured. 
 *
 * SEE ALSO
 *   CreateApplication
 *   DeleteApplication
 *   SelectApplication
 *
 **/
SPROX_API_FUNC(Desfire_GetApplicationIDs) (SPROX_PARAM  BYTE aid_max_count, DWORD aid_list[], BYTE *aid_count)
{
  DWORD   aid;
  BYTE    i;
  WORD    len;
  SPROX_RC   status;
  SPROX_DESFIRE_GET_CTX();

  /* Set the number of applications to zero */
  *aid_count = 0;

  /* create the info block containing the command code */
  ctx->xfer_buffer[INF + 0] = DF_GET_APPLICATION_IDS;
  ctx->xfer_length = 1;

  for (;;)
  {
    status = SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  0, WANTS_ADDITIONAL_FRAME | WANTS_OPERATION_OK);
    if (status != DF_OPERATION_OK)
      return status;

    len = ctx->xfer_length - 1;       /* substract 1 byte for the received status */

    /* ByteCount must be in multiples of APPLICATION_ID_SIZE bytes
       we check this to proof the format integrity
       if zero bytes have been received this is ok -> means no
       applications existing */
    if (len % APPLICATION_ID_SIZE)
      return DFCARD_WRONG_LENGTH;

    for (i = 0; i < (len / APPLICATION_ID_SIZE); i++)
    {
      /* Extract AID */
      aid = ctx->xfer_buffer[INF + 3 + APPLICATION_ID_SIZE * i];
      aid <<= 8;
      aid += ctx->xfer_buffer[INF + 2 + APPLICATION_ID_SIZE * i];
      aid <<= 8;
      aid += ctx->xfer_buffer[INF + 1 + APPLICATION_ID_SIZE * i];

      /* Copy AID */
      aid_list[*aid_count] = aid;

      /* Next AID */
      *aid_count = *aid_count + 1;
      if (*aid_count >= aid_max_count)
      {
        if (i < ((len / APPLICATION_ID_SIZE) - 1))
        {
          /* Still ApplicationIDs available but we ran out
             of buffer space provided by the caller */
          return DFCARD_OVERFLOW;
        }
      }
    }

    /* Prepare the info block indicating that we want more application IDs */
    if (ctx->xfer_buffer[INF + 0] != DF_ADDITIONAL_FRAME)
      return DF_OPERATION_OK;
    ctx->xfer_length = 1;
  }
}

/**f* DesfireAPI/GetDFNames
 *
 * NAME
 *   GetDFNames
 *
 * DESCRIPTION
 *   Returns the Application IDentifiers, ISO DF IDs and ISO DF Names of all active
 *   applications on a DesFire card
 *
 * REMARKS
 *   This function must be called within a loop. Before first call, set follow_up to
 *   FALSE. Loop while follow_up is TRUE.
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_GetDFNames(BOOL  *follow_up,
 *                                  DWORD *aid,
 *                                  WORD  *iso_df_id,
 *                                  BYTE  iso_df_name[16+1],
 *                                  BYTE  *iso_df_namelen);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_GetDFNames(SPROX_INSTANCE rInst,
 *                                   BOOL  *follow_up,
 *                                   DWORD *aid,
 *                                   WORD  *iso_df_id,
 *                                   BYTE   iso_df_name[16+1],
 *                                   BYTE  *iso_df_namelen);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_GetDFNames(SCARDHANDLE hCard,
 *                                 BOOL  *follow_up,
 *                                 DWORD *aid,
 *                                 WORD  *iso_df_id,
 *                                 BYTE   iso_df_name[16+1],
 *                                 BYTE  *iso_df_namelen);
 *
 * INPUTS
 *   BOOL  *follow_up         : IN/OUT : whether a subsequent call is needed, or not
 *   DWORD *aid               : Desfire AID
 *   WORD  *iso_df_id         : ISO ID of the DF
 *   BYTE   iso_df_name[16+1] : ISO name of the DF
 *   BYTE  *iso_df_namelen    : actual length of iso_df_name
 *
 * RETURNS
 *   DF_OPERATION_OK    : operation succeeded
 *   Other code if internal or communication error has occured. 
 *
 * SEE ALSO
 *   CreateApplication
 *   DeleteApplication
 *   SelectApplication
 *
 **/
SPROX_API_FUNC(Desfire_GetDFNames) (SPROX_PARAM  BOOL *follow_up, DWORD *aid, WORD *iso_df_id, BYTE iso_df_name[16+1], BYTE *iso_df_namelen)
{
  SPROX_RC   status;
  SPROX_DESFIRE_GET_CTX();

  if (follow_up == NULL)
    return DFCARD_LIB_CALL_ERROR;

  /* create the info block containing the command code */
  if (*follow_up)
  {
    ctx->xfer_buffer[INF + 0] = DF_ADDITIONAL_FRAME;
    ctx->xfer_length = 1;
  } else
  {
    ctx->xfer_buffer[INF + 0] = DF_GET_DF_NAMES;
    ctx->xfer_length = 1;
  }

  status = SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  0, WANTS_ADDITIONAL_FRAME | WANTS_OPERATION_OK);
  if (status != DF_OPERATION_OK)
    return status;

  if ((ctx->xfer_length < 4) || (ctx->xfer_length > 22))
    return DFCARD_WRONG_LENGTH;

  /* Extract AID */
  if (aid != NULL)
  {
    *aid  = 0;
    *aid += ctx->xfer_buffer[3]; *aid <<= 8;
    *aid += ctx->xfer_buffer[2]; *aid <<= 8;
    *aid += ctx->xfer_buffer[1];
  }

  /* Extract ISO DF ID */
  if (iso_df_id != NULL)
  {
    *iso_df_id  = 0;

    if (ctx->xfer_length > 5)
    {
      *iso_df_id += ctx->xfer_buffer[5]; *iso_df_id <<= 8;
      *iso_df_id += ctx->xfer_buffer[4];
    }
  }

  /* Extract ISO DF Name */
  if (iso_df_name != NULL)
  {
    BYTE i;

    for (i=0; (i<16) && ((6+i)<ctx->xfer_length); i++)
      iso_df_name[i] = ctx->xfer_buffer[6 + i];

    iso_df_name[i] = '\0';

    if (iso_df_namelen != NULL)
      *iso_df_namelen = i;
  }

  *follow_up = (ctx->xfer_buffer[INF + 0] == DF_ADDITIONAL_FRAME) ? TRUE : FALSE;
  return DF_OPERATION_OK;
}

/**f* DesfireAPI/SelectApplication
 *
 * NAME
 *   SelectApplication
 *
 * DESCRIPTION
 *   Selects one specific application for further access
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_SelectApplication(DWORD aid);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_SelectApplication(SPROX_INSTANCE rInst,
 *                                     DWORD aid);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_SelectApplication(SCARDHANDLE hCard,
 *                                     DWORD aid);
 *
 * INPUTS
 *   DWORD aid                   : Application IDentifier
 *
 * RETURNS
 *   DF_OPERATION_OK    : application selected
 *   Other code if internal or communication error has occured. 
 *
 * SEE ALSO
 *   CreateApplication
 *   DeleteApplication
 *   GetApplicationIDs
 *
 **/
SPROX_API_FUNC(Desfire_SelectApplication) (SPROX_PARAM  DWORD aid)
{
  SPROX_DESFIRE_GET_CTX();
  
  /* Each SelectApplication causes a currently valid authentication state to be lost */
  ctx->session_key_id   = KEY_UNSET;
  ctx->session_key_type = KEY_EMPTY;

  /* Create the info block containing the command code and the given parameters. */
  ctx->xfer_buffer[INF + 0] = DF_SELECT_APPLICATION;
  ctx->xfer_buffer[INF + 1] = (BYTE) (aid & 0x000000FF);
  aid >>= 8;
  ctx->xfer_buffer[INF + 2] = (BYTE) (aid & 0x000000FF);
  aid >>= 8;
  ctx->xfer_buffer[INF + 3] = (BYTE) (aid & 0x000000FF);
  ctx->xfer_length = 4;

  /* Communicate the info block to the card and check the operation's return status. */
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  0, WANTS_OPERATION_OK);
}

/**t* DesfireAPI/DF_VERSION_INFO
 *
 * NAME
 *   DF_VERSION_INFO
 *
 * DESCRIPTION
 *   Structure for returning the information supplied by the GetVersion command.
 *
 * SOURCE 
 *   typedef struct
 *   {
 *     // hardware related information
 *     BYTE    bHwVendorID;     // vendor ID (0x04 for NXP)
 *     BYTE    bHwType;         // type (0x01)
 *     BYTE    bHwSubType;      // subtype (0x01)
 *     BYTE    bHwMajorVersion; // major version number
 *     BYTE    bHwMinorVersion; // minor version number
 *     BYTE    bHwStorageSize;  // storage size (0x18 = 4096 bytes)
 *     BYTE    bHwProtocol;     // communication protocol type (0x05 meaning ISO 14443-2 and -3)
 
 *     // software related information
 *     BYTE    bSwVendorID;     // vendor ID (0x04 for NXP)
 *     BYTE    bSwType;         // type (0x01)
 *     BYTE    bSwSubType;      // subtype (0x01)
 *     BYTE    bSwMajorVersion; // major version number
 *     BYTE    bSwMinorVersion; // minor version number
 *     BYTE    bSwStorageSize;  // storage size (0x18 = 4096 bytes)
 *     BYTE    bSwProtocol;     // communication protocol type (0x05 meaning ISO 14443-3 and -4)
 *
 *     BYTE    abUid[7];        // unique serial number
 *     BYTE    abBatchNo[5];    // production batch number
 *     BYTE    bProductionCW;   // calendar week of production
 *     BYTE    bProductionYear; // year of production
 *   } DF_VERSION_INFO;
 *
 **/

/**f* DesfireAPI/GetVersion
 *
 * NAME
 *   GetVersion
 *
 * DESCRIPTION
 *   Returns manufacturing related data of the DesFire card
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_GetVersion(DF_VERSION_INFO *pVersionInfo);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_GetVersion(SPROX_INSTANCE rInst,
 *                                     DF_VERSION_INFO *pVersionInfo);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_GetVersion(SCARDHANDLE hCard,
 *                                     DF_VERSION_INFO *pVersionInfo);
 *
 * INPUTS
 *   DF_VERSION_INFO *pVersionInfo : card's version information
 *
 * RETURNS
 *   DF_OPERATION_OK    : operation succeeded
 *   Other code if internal or communication error has occured. 
 *
 **/
SPROX_API_FUNC(Desfire_GetVersion) (SPROX_PARAM  DF_VERSION_INFO *pVersionInfo)
{
  SPROX_RC   status;
  BYTE   *ptr = (BYTE *) pVersionInfo;
  SPROX_DESFIRE_GET_CTX();

  ctx->xfer_buffer[INF + 0] = DF_GET_VERSION;
  ctx->xfer_length = 1;
  status = SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  8, WANTS_ADDITIONAL_FRAME);
  if (status != DF_OPERATION_OK)
    return status;
  memcpy(ptr, &ctx->xfer_buffer[INF + 1], 7);
  ptr += 7;

  ctx->xfer_buffer[INF + 0] = DF_ADDITIONAL_FRAME;
  ctx->xfer_length = 1;
  status = SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  8, WANTS_ADDITIONAL_FRAME);
  if (status != DF_OPERATION_OK)
    return status;
  memcpy(ptr, &ctx->xfer_buffer[INF + 1], 7);
  ptr += 7;

  ctx->xfer_buffer[INF + 0] = DF_ADDITIONAL_FRAME;
  ctx->xfer_length = 1;
  status = SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  15, WANTS_OPERATION_OK);
  if (status != DF_OPERATION_OK)
    return status;
  memcpy(ptr, &ctx->xfer_buffer[INF + 1], 14);

  return DF_OPERATION_OK;
}



/**f* DesfireAPI/GetIsoFileIDs
 *
 * NAME
 *   GetIsoFileIDs
 *
 * DESCRIPTION
 *   Returns the File IDentifiers of all active files within the currently selected application
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_GetIsoFileIDs(BYTE fid_max_count,
 *                                     WORD fid_list[],
 *                                     BYTE *fid_count);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_GetIsoFileIDs(SPROX_INSTANCE rInst,
 *                                      BYTE fid_max_count,
 *                                      WORD fid_list[],
 *                                      BYTE *fid_count);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_GetIsoFileIDs(SCARDHANDLE hCard,
 *                                    BYTE fid_max_count,
 *                                    WORD fid_list[],
 *                                    BYTE *fid_count);
 *
 * INPUTS
 *   BYTE fid_max_count          : maximum number of File IDentifiers
 *   WORD fid_list[]             : File IDentifiers list 
 *   BYTE *fid_count             : number of File IDentifiers in the selected application
 *
 * RETURNS
 *   DF_OPERATION_OK    : operation succeeded
 *   Other code if internal or communication error has occured. 
 *
 * SEE ALSO
 *   GetFileIDs
 *   GetDfNames
 *
 **/
SPROX_API_FUNC(Desfire_GetIsoFileIDs) (SPROX_PARAM  BYTE fid_max_count, WORD fid_list[], BYTE *fid_count)
{
  SPROX_RC   status;
  SPROX_DESFIRE_GET_CTX();

  /* reset the number of found FileIDs */
  *fid_count = 0;

  /* create the info block containing the command code */
  ctx->xfer_buffer[INF + 0] = DF_GET_ISO_FILE_IDS;
  ctx->xfer_length = 1;

again:

  status = SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  0, WANTS_OPERATION_OK | WANTS_ADDITIONAL_FRAME);
  if (status != DF_OPERATION_OK)
    return status;

  ctx->xfer_length--;                 /* subtract 1 byte for the received status of desfire card */
  for (; (*fid_count < fid_max_count) && (ctx->xfer_length > 0); fid_list++, (*fid_count)++)
  {
    *fid_list  = 0;
    *fid_list += ctx->xfer_buffer[INF + 2 + *fid_count]; *fid_list <<= 8;
    *fid_list += ctx->xfer_buffer[INF + 1 + *fid_count];
    ctx->xfer_length -= 2;
  }

  if (ctx->xfer_buffer[INF + 0] == DF_ADDITIONAL_FRAME)
  {
    ctx->xfer_length = 1;
    goto again;
  }

  /* if there are still bytes left in the ctx->xfer_buffer   */
  /* we ran out of buffer space provided by the caller */
  if (ctx->xfer_length != 0)
    return DFCARD_OVERFLOW;

  return DF_OPERATION_OK;
}





/**f* DesfireAPI/SetConfiguration
 *
 * NAME
 *   SetConfiguration
 *
 * DESCRIPTION
 *   Sends the SetConfiguration command to the DESFIRE card.
 *
 * SYNOPSIS
 *
 *   [[sprox_desfire.dll]]
 *   SWORD SPROX_Desfire_SetConfiguration(BYTE option,
 *                                        const BYTE data[],
 *                                        BYTE length);
 *
 *   [[sprox_desfire_ex.dll]]
 *   SWORD SPROXx_Desfire_SetConfiguration(SPROX_INSTANCE rInst,
 *                                         BYTE option,
 *                                         const BYTE data[],
 *                                         BYTE length);
 *
 *   [[pcsc_desfire.dll]]
 *   LONG  SCardDesfire_SetConfiguration(SCARDHANDLE hCard,
 *                                       BYTE option,
 *                                       const BYTE data[],
 *                                       BYTE length);
 *
 * INPUTS
 *
 * RETURNS
 *   DF_OPERATION_OK : operation succeeded
 *   Other code if internal or communication error has occured. 
 *
 * NOTES
 *   Read DESFIRE EV1 manual, chapter 9.4.9 for details.
 *   DO NOT USE THIS COMMAND unless you're really sure you know
 *   what you're doing.
 *
 **/
SPROX_API_FUNC(Desfire_SetConfiguration) (SPROX_PARAM  BYTE option, const BYTE data[], BYTE length)
{
  DWORD i;
  SPROX_DESFIRE_GET_CTX();

  ctx->xfer_length = 0;

  /* Create the info block containing the command code and the key number argument. */
  ctx->xfer_buffer[ctx->xfer_length++] = DF_SET_CONFIGURATION;
  ctx->xfer_buffer[ctx->xfer_length++] = option;

  if (data != NULL)
  {
    memcpy(&ctx->xfer_buffer[ctx->xfer_length], data, length);
    ctx->xfer_length += length;
  }

  /* Add the CRC */
  Desfire_XferAppendCrc(SPROX_PARAM_P  2);

  printf("\n");
  for (i=0; i<ctx->xfer_length; i++)
    printf("%02X", ctx->xfer_buffer[i]);
  printf("\n");

  /* Cipher the option, the data and the CRC */
  Desfire_XferCipherSend(SPROX_PARAM_P  2);
  
  return SPROX_API_CALL(Desfire_Command) (SPROX_PARAM_P  1, CHECK_RESPONSE_CMAC | WANTS_OPERATION_OK);
}
