#include "cardappz.h"
#include <assert.h>
#include <stdio.h>

/**f* Cardappz/ComputeDivKey
 *
 * NAME
 *   ComputeDivKey
 *
 * DESCRIPTION
 *   Compute a diversified DES or 3-DES key based on a "master key" and card's serial number
 *
 * SYNOPSIS
 *   int KeyDiversification(CARDAPPZ_CTX_ST *ctx, unsigned char DiversifiedKey[16], const unsigned char MasterKey[16]);
 *
 * INPUTS
 *   CARDAPPZ_CTX_ST *ctx              : pointer to the global context
 *   unsigned char DiversifiedKey[16]  : [output] the diversified key
 *   unsigned char MasterKey[16]       : [input]  the master key
 *
 * RETURNS
 *   CARDAPPZ_OK                       : success
 *
 * NOTES
 *   The algorithm is the same as the one implemented in Philips Desfire SAM card (see
 *   Desfire SAM functional specification version 1.0, chapter 4.2.5).
 *   This allow to use indifferently a Desfire SAM or an host-computed key.
 *
 *   - padd card serial number with 0x88 on the left until it is exactly 8 bytes long
 *   - XOR serial number with first part of MasterKey
 *   - apply Triple DES on previous result to get first part of DiversifiedKey
 *   - XOR previous result with second part of MasterKey
 *   - apply Triple DES on previous result to get second part of DiversifiedKey
 *
 *   Both Triple DES operation are encryption with MasterKey.
 *
 **/
int ComputeDivKey(CARDAPPZ_CTX_ST * ctx, unsigned char DiversifiedKey[16],
                  const unsigned char MasterKey[16])
{
  TDES_CTX_ST tdes_ctx;
  unsigned char buffer[8];
  int i;

  if (ctx == NULL)
    return CARDAPPZ_INVALID_CONTEXT;
  if (DiversifiedKey == NULL)
    return CARDAPPZ_INVALID_PARAMETER;
  if (MasterKey == NULL)
    return CARDAPPZ_INVALID_PARAMETER;

  /* Load 3-DES context with master key */
  TDES_Init(&tdes_ctx, &MasterKey[0], &MasterKey[8], &MasterKey[0]);

  /* Preload seed */
  memset(buffer, 0x88, 8);

  /* Put card UID in seed */
  if (ctx->card_type & CARD_TYPE_14443_A)
  {
    if (ctx->card_ident.a.snrlen > 8)
      memcpy(&buffer[0], ctx->card_ident.a.snr, 8);
    else
      memcpy(&buffer[8 - ctx->card_ident.a.snrlen], ctx->card_ident.a.snr,
             ctx->card_ident.a.snrlen);
  } else if (ctx->card_type & CARD_TYPE_14443_B)
  {
    memcpy(&buffer[4], ctx->card_ident.b.atq, 4);
  } else
    return CARDAPPZ_EXCEPTION;

  /* First block <- seed XOR first part of master key */
  for (i = 0; i < 8; i++)
    buffer[i] ^= MasterKey[i];

  /* Cipher first block, this is first part of diversified key */
  TDES_Encrypt(&tdes_ctx, buffer);
  memcpy(&DiversifiedKey[0], buffer, 8);

  /* Second block <- ciphered first block XOR second part of master key */
  for (i = 0; i < 8; i++)
    buffer[i] ^= MasterKey[8 + i];

  /* Cipher second block, this is second part of diversified key */
  TDES_Encrypt(&tdes_ctx, buffer);
  memcpy(&DiversifiedKey[8], buffer, 8);

  return CARDAPPZ_OK;
}


/**f* Cardappz/ComputeSignature
 *
 * NAME
 *   ComputeSignature
 *
 * DESCRIPTION
 *   Compute a static signature for a block of data
 *   The signature algorithm uses an MD5 hash, 2 Triple DES encryption, and a final XOR to build a 8 bytes Message Authentication Code (MAC)
 *
 * SYNOPSIS
 *   int ComputeSignature(CARDAPPZ_CTX_ST *ctx, unsigned char Signature[8], unsigned char Data[], unsigned long Size, const unsigned char SignKey[16]);
 *
 * INPUTS
 *   CARDAPPZ_CTX_ST *ctx              : pointer to the global context
 *   unsigned char Signature[8]        : [output] the computed MAC
 *   unsigned char Data[]              : [input]  the data to be signed
 *   unsigned long Size                : size of the data
 *   unsigned char SignKey[16]         : [input]  the signature key
 *
 * RETURNS
 *   CARDAPPZ_OK                       : success
 *
 * NOTES
 *   Here's the algorithm :
 *
 *   - compute MD5 of card serial number followed by Data
 *   - apply Triple DES on the first part of the MD5 to get a first block
 *   - apply Triple DES on the second part of the MD5 to get a second block
 *   - XOR both first and second block to get Signature
 *
 *   Both Triple DES operation are encryption with SignKey.
 *
 **/
int ComputeSignature(CARDAPPZ_CTX_ST * ctx, unsigned char Signature[8],
                     unsigned char Data[], unsigned long Size,
                     const unsigned char SignKey[16])
{
  MD5_CTX_ST md5_ctx;
  TDES_CTX_ST tdes_ctx;
  unsigned char digest[16];
  int i;

  if (ctx == NULL)
    return CARDAPPZ_INVALID_CONTEXT;
  if (SignKey == NULL)
    return CARDAPPZ_INVALID_PARAMETER;
  if (Signature == NULL)
    return CARDAPPZ_INVALID_PARAMETER;
  if (Data == NULL)
    return CARDAPPZ_INVALID_PARAMETER;
  if (Size == 0)
    return CARDAPPZ_INVALID_PARAMETER;

  /* Load 3-DES context with sigature key */
  TDES_Init(&tdes_ctx, &SignKey[0], &SignKey[8], &SignKey[0]);

  /* Init MD5 operation */
  MD5_Init(&md5_ctx);

  /* Hash card UID */
  if (ctx->card_type & CARD_TYPE_14443_A)
  {
    MD5_Update(&md5_ctx, ctx->card_ident.a.snr, ctx->card_ident.a.snrlen);
  } else if (ctx->card_type & CARD_TYPE_14443_B)
  {
    MD5_Update(&md5_ctx, ctx->card_ident.b.atq, 4);
  } else
    return CARDAPPZ_EXCEPTION;


  /* Hash data */
  MD5_Update(&md5_ctx, Data, Size);

  /* Get hash result */
  MD5_Final(&md5_ctx, digest);

  /* Cipher both blocks */
  TDES_Encrypt(&tdes_ctx, &digest[0]);
  TDES_Encrypt(&tdes_ctx, &digest[8]);

  /* Compute signature as XOR of both blocks */
  for (i = 0; i < 8; i++)
    Signature[i] = digest[i] ^ digest[8 + i];

  /* Done */
  return CARDAPPZ_OK;
}
