#include <vcl.h>
#pragma hdrstop

#include "uWorker.h"
#include "uMain.h"
#include "uProgress.h"

#include "msoilv.h"
#include "jaycos.h"

#include "springprox.h"
#include "cardware/desfire/sprox_desfire.h"

#include <stdio.h>

#pragma package(smart_init)

bool read_file(const char *file_name, unsigned char *buffer, size_t max_size,
               unsigned long *actual_size)
{
  size_t rc;

  if (file_name == NULL)
    return false;
  if (buffer == NULL)
    return false;
  FILE *fp = fopen(file_name, "rb");

  if (fp == NULL)
    return false;
  rc = fread(buffer, 1, max_size, fp);
  fclose(fp);
  if (rc <= 0)
    return false;
  if (actual_size != NULL)
    *actual_size = rc;
  return true;
}

bool write_file(const char *file_name, unsigned char *buffer, size_t size)
{
  size_t rc;

  if (file_name == NULL)
    return false;
  if (buffer == NULL)
    return false;
  FILE *fp = fopen(file_name, "wb");

  if (fp == NULL)
    return false;
  rc = fwrite(buffer, 1, size, fp);
  fclose(fp);
  return (rc == size);
}

__fastcall TWorker::TWorker(bool CreateSuspended):TThread(CreateSuspended)
{
  bReaderActive = false;
  hWakeupEvent = CreateEvent(NULL, false, false, NULL);
}

bool __fastcall TWorker::OpenReaderTask()
{
  short rc;

  if (bReaderActive)
    return true;

  /* Display the message */
  WriteLn("Connecting to SpringProx reader...");

  bReaderHasFingerprint = false;

  rc = SPROX_ReaderOpen(NULL);
  if (rc == MI_OK)
  {
    /* Reset the RF field */
    SPROX_ControlRF(5);
    /* OK, great ! */
    WriteLn("Success !");
    bReaderActive = true;

    /* Check if we have a fingerprint reader or not */
    DWORD features = 0;

    if (SPROX_ReaderGetFeatures(&features) == MI_OK)
      if (features & 0x00000040)
        bReaderHasFingerprint = true;

  } else
    WriteLn("Reader not found.");


  return bReaderActive;
}

bool __fastcall TWorker::FindCardTask()
{
  short rc;

  if (!bReaderActive)
  {
    WriteLn("Not connected to reader");
    return false;
  }

  /* We must halt the card first (in case we were already connected) */
  if (fMain->card_data.is_tcl)
  {
    SPROX_Tcl_Deselect(0xFF);
  } else if (fMain->card_data.is_type_a)
  {
    SPROX_A_Halt();
  } else if (fMain->card_data.is_type_b)
  {
    SPROX_B_Halt(fMain->card_data.iso.b.atq);
  }

  WriteLn("Looking for a card...");

  fMain->card_data.is_type_a = false;
  fMain->card_data.is_type_b = false;
  fMain->card_data.is_tcl = false;
  fMain->card_data.is_type = CARD_UNKNOWN;

  /* At first try to find an ISO-14443-A card */
  SPROX_SetConfig(CFG_MODE_ISO_14443_A);

  fMain->card_data.iso.a.snrlen = sizeof(fMain->card_data.iso.a.snr);
  rc =
    SPROX_A_SelectAny(fMain->card_data.iso.a.atq, fMain->card_data.iso.a.snr,
                      &fMain->card_data.iso.a.snrlen,
                      fMain->card_data.iso.a.sak);

  if (rc == MI_OK)
  {
    /* A card has been found */
    WriteLn("Type A card found");

    if (fMain->card_data.iso.a.sak[0] & 0x20)
    {
      /* Card support T=CL, try to activate it */
      fMain->card_data.tcl.atslen = sizeof(fMain->card_data.tcl.ats);
      rc =
        SPROX_TclA_GetAts(0xFF, fMain->card_data.tcl.ats,
                          &fMain->card_data.tcl.atslen);
      if (rc == MI_OK)
      {
        /* Found a type A T=CL card */
        fMain->card_data.is_type_a = true;
        fMain->card_data.is_tcl = true;
        WriteLn("T=CL card activated");

        /* Recognize the Desfire from its ATQ */
        if ((fMain->card_data.iso.a.atq[1] == 0x03)
            && (fMain->card_data.iso.a.atq[0] == 0x44))
          fMain->card_data.is_type = CARD_DESFIRE;

        return true;
      }
    } else
    {
      /* No T=CL support, but type A card found anyway */
      fMain->card_data.is_type_a = true;

      /* Recognize the card family from its ATQ */
      if ((fMain->card_data.iso.a.atq[1] == 0x00)
          && (fMain->card_data.iso.a.atq[0] == 0x04))
      {
        fMain->card_data.is_type = CARD_MIFARE_1K;
      } else
        if ((fMain->card_data.iso.a.atq[1] == 0x00)
            && (fMain->card_data.iso.a.atq[0] == 0x02))
      {
        fMain->card_data.is_type = CARD_MIFARE_4K;
      } else
        if ((fMain->card_data.iso.a.atq[1] == 0x03)
            && (fMain->card_data.iso.a.atq[0] == 0x40))
      {
        fMain->card_data.is_type = CARD_MIFARE_UL;
      }

      return true;
    }

    /* Type A card found but with error */
    WriteLn("Activation error");
    WriteRc(rc);
    return false;
  }

  /* Nox try to find an ISO-14443-B card */
  SPROX_SetConfig(CFG_MODE_ISO_14443_B);

  rc = SPROX_B_SelectAny(0x00, fMain->card_data.iso.b.atq);

  if (rc == MI_OK)
  {
    /* A card has been found */
    WriteLn("Type B card found");

    if (fMain->card_data.iso.b.atq[10] & 0x01)
    {
      /* Card support T=CL, try to activate it */
      fMain->card_data.tcl.atslen = 0;
      rc = SPROX_TclB_Attrib(fMain->card_data.iso.b.atq, 0xFF);
      if (rc == MI_OK)
      {
        /* Found a type B T=CL card */
        fMain->card_data.is_type_b = true;
        fMain->card_data.is_tcl = true;
        WriteLn("T=CL card activated");
        return true;
      }
    } else
    {
      /* No T=CL support, but type Bcard found anyway */
      fMain->card_data.is_type_b = true;
      return true;
    }

    /* Type B card found but with error */
    WriteLn("Activation error");
    WriteRc(rc);
    return false;
  }

  WriteLn("No card on the reader (or card not available)");
  return false;
}

/*
 ****************************************************************************
 *
 *  FINGERPRINTS
 *
 ****************************************************************************
 */

void __fastcall TWorker::EnrollAsyncAction(unsigned char code)
{
  switch (code)
  {
    case 0:
      /* WriteLn("Place finger"); */
      break;
    case 1:
      WriteLn("Move up");
      break;
    case 2:
      WriteLn("Move down");
      break;
    case 3:
      WriteLn("Move left");
      break;
    case 4:
      WriteLn("Move right");
      break;
    case 5:
      WriteLn("Press harder");
      break;
    case 6:
      WriteLn("Move finger");
      break;
    case 7:
      WriteLn("Remove finger");
      break;
    case 8:
      WriteLn("Finger OK");
      break;
  }
}

void __fastcall TWorker::EnrollAsyncInfo(unsigned char finger_no,
                                         unsigned char finger_count,
                                         unsigned char capture_no,
                                         unsigned char capture_count)
{
  char buffer[128];

  sprintf(buffer, "Finger %d/%d, capture %d/%d", finger_no, finger_count,
          capture_no, capture_count);
  WriteLn(buffer);
}

bool __fastcall TWorker::EnrollTask()
{
  BYTE buffer[1024];
  BYTE async;
  WORD size;
  BYTE tag;
  WORD offset, length;
  BYTE *value;
  BYTE idx;
  SWORD rc;

  fMain->card_data.finger_size[0] = 0;
  fMain->card_data.finger_size[1] = 0;
  memset(fMain->card_data.finger_data, 0, FINGERS_SIZE);

  /* Build the ILV enroll command               */
  /* (SEE SAGEM DOCUMENTATION FOR EXPLANATIONS) */
  /* ------------------------------------------ */

  memset(buffer, 0, sizeof(buffer));
  length = 0;

  /* Initialize ILV */
  ilv_init(buffer, &length, ILV_ENROLL);

  /* Build ILV */
  ilv_add_byte(buffer, &length, 0); /* Database identifier = 0 */
  ilv_add_word(buffer, &length, 60);  /* Timeout = 60 seconds */
  ilv_add_byte(buffer, &length, 0); /* Acquisition quality threshold = default (0) */
  ilv_add_byte(buffer, &length, 0); /* Enrollment type = default (0) */
  ilv_add_byte(buffer, &length, 2); /* We want to enroll 2 fingers */
  ilv_add_byte(buffer, &length, 0); /* Save record into local database = NO (0) */
  ilv_add_byte(buffer, &length, 1); /* Export minutiae with its default size = YES (1) */

  /* We want to receive asynchronous events */
  ilv_add_byte(buffer, &length, ID_ASYNCHRONOUS_EVENT);
  ilv_add_word(buffer, &length, 4);
  ilv_add_dword(buffer, &length, 0x00000005);
  /* Bit field = 0x00000005 : action messages (move left, press harder...) */
  /* and information messages (finger X/X, capture X/X) will be sent       */

  /* Write length in the ILV */
  ilv_final(buffer, &length);

  /* Send the command to the MSO and process its responses (including asynchronous messages) */
  /* --------------------------------------------------------------------------------------- */
  async = FALSE;
  for (;;)
  {
    size = sizeof(buffer);

    /* NOTE : the timeout of the exchange must be a little greater than the timeout offered  */
    /*        to the MSO, overwhise the library will return an error code before the MSO has */
    /*        finished is process                                                            */
    if (async)
      rc = SPROX_MSO_Exchange(NULL, 0, buffer, &size, 65000, &async);
    else
      rc = SPROX_MSO_Exchange(buffer, length, buffer, &size, 65000, &async);

    if (rc != 0)
      return false;

    if (async)
    {
      /* Easy processing of asynchronous message, no error control */
      switch (buffer[4])
      {
        case 0x01:
          EnrollAsyncAction(buffer[7]);
          break;
        case 0x04:
          EnrollAsyncInfo(buffer[7], buffer[8], buffer[9], buffer[10]);
          break;
      }

      /* NOTE : we MUST go back to MSO_Exchange AS SOON AS POSSIBLE because if we were to  */
      /*        miss any message (asynchronous or final result), the reader will report a  */
      /*        protocol error and abort the process                                       */
    } else
    {
      break;                    /* Last message -> end of the loop */
    }
  }

  /* 
   * Reply format is :
   *  0      : reply ID
   *  1 -> 2 : reply length
   *  3      : request status
   *  4      : enroll status
   *  5 -> 8 : user database index
   *  9      : first finger
   *  ...    : second finger
   */

  /* Check returned reply and status */
  /* ------------------------------- */

  if (buffer[0] != ILV_ENROLL)
  {
    WriteLn("Enroll failed : MSO reply=0x" + IntToHex(buffer[0], 2));
    return false;
  }
  if (buffer[3] != 0x00)
  {
    WriteLn("Enroll failed : MSO request status=0x" + IntToHex(buffer[3], 2));
    return false;
  }
  if (buffer[4] != 0x00)
  {
    WriteLn("Enroll failed : MSO enroll status=0x" + IntToHex(buffer[4], 2));
    return false;
  }

  /* Decode returned data */
  /* -------------------- */

  offset = 9;
  idx = 0;
  while ((offset < size) && ilv_loop(buffer, &offset, &tag, &length, &value))
  {
    if ((tag != ID_PKCOMP)      /* PKCOMP minutiae */
        && (tag != ID_PKMAT))   /* PKMAT minutiae */
    {
      /* Skip */
      continue;
    }

    if (idx < 2)
    {
      if (length <= FINGERS_SIZE / 2)
      {
        fMain->card_data.finger_size[idx] = length;
        fMain->card_data.finger_type = tag; /* Both fingerprints use the same format ! */
        memcpy(fMain->card_data.finger_data[idx], value, length);
      } else
      {
        WriteLn("Finger's minutiae are too big to be stored in a card");
      }
      idx++;
    } else
    {
      /* Silently ignore the finger */
    }
  }

  return (fMain->card_data.finger_size[0] || fMain->card_data.finger_size[1]);
}

void __fastcall TWorker::VerifyAsyncAction(unsigned char code)
{
  switch (code)
  {
    case 0:
      WriteLn("Place finger");
      break;
    default:
      EnrollAsyncAction(code);
  }
}
bool __fastcall TWorker::VerifyTask()
{
  BYTE buffer[1024];
  BYTE async;
  WORD size;
  WORD length;
  SWORD rc;

  /* Build the ILV verify command               */
  /* (SEE SAGEM DOCUMENTATION FOR EXPLANATIONS) */
  /* ------------------------------------------ */

  memset(buffer, 0, sizeof(buffer));
  length = 0;

  /* Initialize ILV */
  ilv_init(buffer, &length, ILV_VERIFY);

  /* Build ILV */
  ilv_add_word(buffer, &length, 30);  /* Timeout = 60 seconds */
  ilv_add_word(buffer, &length, 5); /* Matching threshold = 5 (recommended value) */
  ilv_add_byte(buffer, &length, 0); /* Acquisition quality threshold = default (0) */

  /* Add first finger minutiae */
  if (fMain->card_data.finger_size[0])
  {
    ilv_add_byte(buffer, &length, fMain->card_data.finger_type);
    ilv_add_word(buffer, &length, fMain->card_data.finger_size[0]);
    ilv_add_data(buffer, &length, fMain->card_data.finger_data[0],
                 fMain->card_data.finger_size[0]);
  }

  /* Add second finger minutiae */
  if (fMain->card_data.finger_size[1])
  {
    ilv_add_byte(buffer, &length, fMain->card_data.finger_type);
    ilv_add_word(buffer, &length, fMain->card_data.finger_size[1]);
    ilv_add_data(buffer, &length, fMain->card_data.finger_data[1],
                 fMain->card_data.finger_size[1]);
  }

  /* We want to receive asynchronous events */
  ilv_add_byte(buffer, &length, ID_ASYNCHRONOUS_EVENT);
  ilv_add_word(buffer, &length, 4);
  ilv_add_dword(buffer, &length, 0x00000001);

  /* Write length in the ILV */
  ilv_final(buffer, &length);

  /* Send the command to the MSO and process its responses (including asynchronous messages) */
  /* --------------------------------------------------------------------------------------- */
  async = FALSE;
  for (;;)
  {
    size = sizeof(buffer);

    /* NOTE : the timeout of the exchange must be a little greater than the timeout offered  */
    /*        to the MSO, overwhise the library will return an error code before the MSO has */
    /*        finished is process                                                            */
    if (async)
      rc = SPROX_MSO_Exchange(NULL, 0, buffer, &size, 35000, &async);
    else
      rc = SPROX_MSO_Exchange(buffer, length, buffer, &size, 35000, &async);

    if (rc != 0)
      return 0;                 /* Error */

    if (async)
    {
      /* Easy processing of asynchronous message, no error control */
      switch (buffer[4])
      {
        case 0x01:
          VerifyAsyncAction(buffer[7]);
          break;
      }

      /* NOTE : we MUST go back to MSO_Exchange AS SOON AS POSSIBLE because if we were to  */
      /*        miss any message (asynchronous or final result), the reader will report a  */
      /*        protocol error and abort the process                                       */
    } else
    {
      break;                    /* Last message -> end of the loop */
    }
  }

  /* 
   * Reply format is :
   *  0      : reply ID
   *  1 -> 2 : reply length
   *  3      : request status
   *  4      : matching result
   */

  /* Check returned reply and status */
  /* ------------------------------- */

  if (buffer[0] != ILV_VERIFY)
  {
    WriteLn("Verify failed : MSO reply=0x" + IntToHex(buffer[0], 2));
    return 0;
  }
  if (buffer[3] != 0x00)
  {
    WriteLn("Verify failed : MSO request status=0x" + IntToHex(buffer[3], 2));
    return 0;
  }

  /* Check matching result */
  /* --------------------- */

  switch (buffer[4])
  {
    case ILVSTS_HIT:           /* Verify -> OK ! */
      return true;

    case ILVSTS_NO_HIT:        /* Verify -> KO ! */
      WriteLn("No match !!!");
      return false;
  }

  WriteLn("Verify failed : MSO matching result=0x" + IntToHex(buffer[4], 2));
  return false;
}

/*
 ****************************************************************************
 *
 *  DESFIRE
 *
 ****************************************************************************
 */

#define DESFIRE_APPLICATION_ID   0x00112233
#define DESFIRE_IDENT_FILE_ID    0
#define DESFIRE_PICTURE_FILE_ID  1
#define DESFIRE_FINGER_FILE_ID   2
#define DESFIRE_BALANCE_FILE_ID  3
#define DESFIRE_HISTORY_FILE_ID  4

static const unsigned char DESFIRE_BLANK_KEY[16] =
  { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x00, 0x00
};
static const unsigned char DESFIRE_KEY_0[16] =
  { 0xB0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x11, 0x22, 0x33
};
static const unsigned char DESFIRE_KEY_1[16] =
  { 0xA0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x11, 0x22, 0x33
};

bool __fastcall TWorker::ReadCardDesfire()
{
  unsigned char buffer[4096];
  unsigned long length;
  int i;
  short rc;

  WriteLn("Reading Desfire card");

  /* Select our application in the Desfire card */
  WriteLn("Entering application directory");
  rc = DesFireCard_SelectApplication(DESFIRE_APPLICATION_ID);
  if (rc != MI_OK)
  {
    WriteLn("Card application not found");
    WriteRc(rc);
    return false;
  }

  /* Get authenticated with key 1 */
  rc = DesFireCard_Authenticate(1, DESFIRE_KEY_1);
  if (rc != MI_OK)
  {
    WriteLn("Authentication failed !");
    WriteRc(rc);
    return false;
  }

  /* Read the information file. This one is mandatory */
  WriteLn("Reading the information block");
  rc = DesFireCard_ReadData2(DESFIRE_IDENT_FILE_ID, 0, 0, buffer, &length);
  if (rc != MI_OK)
  {
    WriteLn("Read from card failed");
    WriteRc(rc);
    return false;
  }

  /* Translate information file into interesting content... */
  DecodeUserData(buffer);

  /* Read the picture (if some) */
  if (fMain->card_data.picture_size)
  {
    WriteLn("Reading the picture");
    rc =
      DesFireCard_ReadData2(DESFIRE_PICTURE_FILE_ID, 0, 0, buffer, &length);
    if (rc == MI_OK)
    {
      /* Truncate to specified size */
      if (fMain->card_data.picture_size > length)
        fMain->card_data.picture_size = length;
      /* Save the picture into a temporary file */
      if (!write_file
          (fMain->temp_picture_file_name.c_str(), buffer,
           fMain->card_data.picture_size))
      {
        WriteLn("Failed to save picture to temporary file");
        WriteRc(rc);
        return false;
      }
      fMain->card_data.picture_file_name = fMain->temp_picture_file_name;
    } else
    {
      WriteLn("Read from card failed");
      fMain->card_data.picture_size = 0;
      WriteRc(rc);
      return false;
    }
  }

  /* Read the fingerprint file (if some) */
  if (fMain->card_data.finger_size[0] || fMain->card_data.finger_size[1])
  {
    WriteLn("Reading the fingerprints");
    memset(buffer, 0, sizeof(buffer));
    rc = DesFireCard_ReadData2(DESFIRE_FINGER_FILE_ID, 0, 0, buffer, &length);
    if (rc == MI_OK)
    {
      memcpy(fMain->card_data.finger_data, buffer, FINGERS_SIZE);
    } else
    {
      WriteLn("Read from card failed");
      fMain->card_data.finger_size[0] = 0;
      fMain->card_data.finger_size[1] = 0;
      WriteRc(rc);
      return false;
    }
  }

  return true;
}

bool __fastcall TWorker::WriteCardDesfire()
{
  unsigned char buffer[4096];
  unsigned long length;
  short rc;

  WriteLn("Writing Desfire card");

  /* Select our application in the Desfire card */
  WriteLn("Entering application directory");
  rc = DesFireCard_SelectApplication(DESFIRE_APPLICATION_ID);
  if (rc != MI_OK)
  {
    /* Application doesn't exist, let's create it ! */
    if (!FormatCardDesfire())
      return false;

    /* Assume we're still in application directory */
  }

  /* Get authenticated with key 0 */
  rc = DesFireCard_Authenticate(0, DESFIRE_KEY_0);
  if (rc != MI_OK)
  {
    WriteLn("Authentication failed !");
    WriteRc(rc);
    return false;
  }

  /* Write the picture file */
  if (fMain->card_data.picture_size)
  {
    if (!read_file
        (fMain->card_data.picture_file_name.c_str(), buffer, sizeof(buffer),
         &length))
    {
      WriteLn("Failed to load the picture file");
      return false;
    }
    if (length > PICTURE_SIZE)
      fMain->card_data.picture_size = PICTURE_SIZE;
    else
      fMain->card_data.picture_size = length;
    WriteLn("Writing the picture");
    rc =
      DesFireCard_WriteData2(DESFIRE_PICTURE_FILE_ID, 0,
                             fMain->card_data.picture_size, buffer);
    if (rc != MI_OK)
    {
      WriteLn("Write to card failed");
      WriteRc(rc);
      return false;
    }
  }

  /* Write the fingerprint file */
  if (fMain->card_data.finger_size[0] || fMain->card_data.finger_size[1])
  {
    WriteLn("Writing the fingerprints");
    rc =
      DesFireCard_WriteData2(DESFIRE_FINGER_FILE_ID, 0, FINGERS_SIZE,
                             &fMain->card_data.finger_data[0][0]);
    if (rc != MI_OK)
    {
      WriteLn("Write to card failed");
      WriteRc(rc);
      return false;
    }
  }

  /* Build the information file content */
  EncodeUserData(buffer);

  /* Write the information file */
  WriteLn("Writing the information block");
  rc = DesFireCard_WriteData2(DESFIRE_IDENT_FILE_ID, 0, IDENT_SIZE, buffer);
  if (rc != MI_OK)
  {
    WriteLn("Write to card failed");
    WriteRc(rc);
    return false;
  }

  return true;
}

bool __fastcall TWorker::FormatCardDesfire()
{
  unsigned long length;
  short rc;

  WriteLn("Formatting Desfire card");

  /* Select root application in the Desfire card */
  rc = DesFireCard_SelectApplication(0);
  if (rc != MI_OK)
  {
    WriteLn("Card fatal error");
    WriteRc(rc);
    return false;
  }

  /* Perform 3-DES authentication with card's master key (must be blank !) */
  rc = DesFireCard_Authenticate(0, DESFIRE_BLANK_KEY);
  if (rc != MI_OK)
  {
    WriteLn("Card authentication failed !");
    WriteRc(rc);
    return false;
  }

  /* Create our application. Application has 2 keys. We don't care of key_settings here. */
  WriteLn("Creating Desfire application");
  rc = DesFireCard_CreateApplication(DESFIRE_APPLICATION_ID, 0xE3, 2);
  if (rc != MI_OK)
  {
    WriteLn("Failed to create the card application");
    WriteRc(rc);
    return false;
  }

  /* Select the newly created application */
  rc = DesFireCard_SelectApplication(DESFIRE_APPLICATION_ID);
  if (rc != MI_OK)
  {
    WriteLn("Failed to select the card application");
    WriteRc(rc);
    return false;
  }

  /* Get authenticated with blank key 0 */
  rc = DesFireCard_Authenticate(0, DESFIRE_BLANK_KEY);
  if (rc != MI_OK)
  {
    WriteLn("Authentication failed !");
    WriteRc(rc);
    return false;
  }

  /* Create the files */
  WriteLn("Creating files inside card application");
  rc =
    DesFireCard_CreateStdDataFile(DESFIRE_IDENT_FILE_ID, DF_COMM_MODE_PLAIN,
                                  0x1F00, IDENT_SIZE);
  if (rc != MI_OK)
    goto failed;
  rc =
    DesFireCard_CreateStdDataFile(DESFIRE_PICTURE_FILE_ID, DF_COMM_MODE_PLAIN,
                                  0x1F00, PICTURE_SIZE);
  if (rc != MI_OK)
    goto failed;
  rc =
    DesFireCard_CreateStdDataFile(DESFIRE_FINGER_FILE_ID,
                                  DF_COMM_MODE_ENCIPHERED, 0x1F00,
                                  FINGERS_SIZE);
  if (rc != MI_OK)
    goto failed;
  rc =
    DesFireCard_CreateValueFile(DESFIRE_BALANCE_FILE_ID, DF_COMM_MODE_MACED,
                                0x1F00, 0, 100000, 0, false);
  if (rc != MI_OK)
    goto failed;
  rc =
    DesFireCard_CreateCyclicRecordFile(DESFIRE_HISTORY_FILE_ID,
                                       DF_COMM_MODE_PLAIN, 0xEEEE, 8, 4);
  if (rc != MI_OK)
    goto failed;

  /* Put the new keys */
  WriteLn("Installing new keys");
  rc = DesFireCard_ChangeKey(0, DESFIRE_KEY_0, NULL);
  if (rc != MI_OK)
  {
    WriteLn("Failed to install Key #0 !");
    WriteRc(rc);
    return false;
  }

  /* Get authenticated with blank key 1 */
  rc = DesFireCard_Authenticate(1, DESFIRE_BLANK_KEY);
  if (rc != MI_OK)
  {
    WriteLn("Authentication failed !");
    WriteRc(rc);
    return false;
  }
  rc = DesFireCard_ChangeKey(1, DESFIRE_KEY_1, NULL);
  if (rc != MI_OK)
  {
    WriteLn("Failed to install Key #1 !");
    WriteRc(rc);
    return false;
  }

  /* Done ! */
  return true;

failed:
  WriteLn("Failed to create on or more file in the card");
  WriteRc(rc);
  return false;
}

bool __fastcall TWorker::EraseCardDesfire()
{
  unsigned char buffer[4096];
  unsigned long length;
  short rc;

  WriteLn("Erasing Desfire card");

  /* Select root application in the Desfire card */
  rc = DesFireCard_SelectApplication(0);
  if (rc != MI_OK)
  {
    WriteLn("Card fatal error");
    WriteRc(rc);
    return false;
  }

  /* Perform 3-DES authentication with card's master key (must be blank !) */
  rc = DesFireCard_Authenticate(0, DESFIRE_BLANK_KEY);
  if (rc != MI_OK)
  {
    WriteLn("Card authentication failed !");
    WriteRc(rc);
    return false;
  }

  /* OK, now let's format the card. Remember this destroys everything (and not only our application) */
  rc = DesFireCard_FormatPICC();
  if (rc != MI_OK)
  {
    WriteLn("Failed to erase the card");
    WriteRc(rc);
    return false;
  }

  /* Done ! */
  return true;
}

/*
 ****************************************************************************
 *
 *  MIFARE 1K AND 4K
 *
 ****************************************************************************
 */

static const unsigned char MIFARE_HEADER[16] =
  { 0x00, 0x00, 'P', 'r', 'o', '-', 'A', 'c', 't', 'i', 'v', 'e', 0x00, 0x11,
  0x22, 0x33
};
static const unsigned char MIFARE_KEY_A[6] =
  { 0xA0, 0xA1, 0xA2, 0xA3, 0xA4, 0xA5 };
static const unsigned char MIFARE_KEY_B[6] =
  { 0x50, 0x41, 0x00, 0x11, 0x22, 0x33 };

/* Mifare transport key - The erase and format functions work only if this key */

/* is in the key set of the reader (shoul'd be in internal EEPROM)             */
static const unsigned char MIFARE_TRANSPORT_KEY[6] =
  { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF };

bool __fastcall TWorker::ReadCardMifare1k()
{
  unsigned char buffer[11 * 48];
  unsigned char sector;
  int i;
  short rc;

  WriteLn("Reading Mifare card");

  for (sector = 0; sector < 11; sector++)
  {
    rc =
      SPROX_MifStReadSector(NULL, sector, &buffer[48 * sector], MIFARE_KEY_A);
    if (rc != MI_OK)
    {
      WriteLn("Failed to read sector " + IntToStr(sector));
      WriteRc(rc);
      return false;
    }

    if (sector == 0)
    {
      /* Check that the expected header is found on blocks 1 & 2 */
      if (memcmp(&buffer[16], MIFARE_HEADER, 16))
      {
        WriteLn("Card doesn't have our header");
        return false;
      }
    }
  }

  /* Translate information block into interesting content... */
  /* Information block starts at offset 48 */
  DecodeUserData(&buffer[48]);

  /* Read the fingerprint file (if some) */
  /* Fingerprint block starts at offset 128 */
  if (fMain->card_data.finger_size[0] || fMain->card_data.finger_size[1])
  {
    memcpy(fMain->card_data.finger_data, &buffer[128], FINGERS_SIZE);
  }

  return true;
}

bool __fastcall TWorker::ReadCardMifare4k()
{
  unsigned char buffer[PICTURE_SIZE];
  unsigned char sector;
  unsigned short length;
  int i;
  short rc;

  /* First part is the same as Mifare 1k */
  if (!ReadCardMifare1k())
    return false;

  /* The second part is only the picture */
  if (fMain->card_data.picture_size)
  {
    WriteLn("Reading the picture");

    length = 0;
    for (sector = 11; sector < 40; sector++)
    {
      rc = SPROX_MifStReadSector(NULL, sector, &buffer[length], MIFARE_KEY_A);
      if (rc != MI_OK)
      {
        WriteLn("Failed to read sector " + IntToStr(sector));
        WriteRc(rc);
        fMain->card_data.picture_size = 0;
        return false;
      }

      if (sector < 32)
      {
        /* Short sector */
        length += 48;
      } else
      {
        /* Long sector */
        length += 240;
      }

      if (length >= fMain->card_data.picture_size)
        break;                  /* Done ! */
    }

    /* Save the picture into a temporary file */
    if (!write_file
        (fMain->temp_picture_file_name.c_str(), buffer,
         fMain->card_data.picture_size))
    {
      WriteLn("Failed to save picture to temporary file");
      WriteRc(rc);
      return false;
    }
    fMain->card_data.picture_file_name = fMain->temp_picture_file_name;
  }

  return true;
}

bool __fastcall TWorker::WriteCardMifare1k()
{
  unsigned char buffer[11 * 48];
  unsigned char sector;
  int i;
  short rc;

  memset(buffer, 0, sizeof(buffer));

  /* Check that sector 0 has our marker and our secret key */
  rc = SPROX_MifStReadSector(NULL, 0, &buffer[0], MIFARE_KEY_B);
  if ((rc != MI_OK) || (memcmp(&buffer[16], MIFARE_HEADER, 16)))
  {
    /* The card must be formated */
    if (!FormatCardMifare())
      return false;
  }

  /* Prepare the fingerprint block */
  if (fMain->card_data.finger_size[0] || fMain->card_data.finger_size[1])
  {
    memcpy(&buffer[128], fMain->card_data.finger_data, FINGERS_SIZE);
  }

  /* Build the information file content */
  EncodeUserData(&buffer[48]);

  /* Now writing */
  WriteLn("Writing Mifare card");
  for (sector = 1; sector < 11; sector++)
  {
    rc =
      SPROX_MifStWriteSector(NULL, sector, &buffer[48 * sector],
                             MIFARE_KEY_B);
    if (rc != MI_OK)
    {
      WriteLn("Failed to write sector " + IntToStr(sector));
      WriteRc(rc);
      return false;
    }
  }

  return true;
}

bool __fastcall TWorker::WriteCardMifare4k()
{
  unsigned char buffer[PICTURE_SIZE];
  unsigned char sector;
  unsigned long length;
  int i;
  short rc;

  /* We must retrieve the picture size before actually writing the information block */
  if (fMain->card_data.picture_size)
  {
    if (!read_file
        (fMain->card_data.picture_file_name.c_str(), buffer, sizeof(buffer),
         &length))
    {
      WriteLn("Failed to load the picture file");
      return false;
    }

    if (length > PICTURE_SIZE)
      fMain->card_data.picture_size = PICTURE_SIZE;
    else
      fMain->card_data.picture_size = length;
  }

  /* First part is the same as Mifare 1k */
  if (!WriteCardMifare1k())
    return false;

  /* The second part is only the picture */
  if (fMain->card_data.picture_size)
  {
    WriteLn("Writing the picture to the card");
    length = 0;
    for (sector = 11; sector < 40; sector++)
    {
      rc =
        SPROX_MifStWriteSector(NULL, sector, &buffer[length], MIFARE_KEY_B);
      if (rc != MI_OK)
      {
        WriteLn("Failed to write sector " + IntToStr(sector));
        WriteRc(rc);
        return false;
      }

      if (sector < 32)
      {
        /* Short sector */
        length += 48;
      } else
      {
        /* Long sector */
        length += 240;
      }

      if (length >= fMain->card_data.picture_size)
        break;                  /* Done ! */
    }
  }

  return true;
}

bool __fastcall TWorker::FormatCardMifare()
{
  unsigned char buffer[48];
  unsigned char sector;
  short rc;

  WriteLn("Formatting Mifare card");
  WriteLn("... be patient, this may take a long time ...");

  /* Put our keys in each sector */
  for (sector = 0; sector < 40; sector++)
  {
    if ((sector >= 16) && (fMain->card_data.is_type != CARD_MIFARE_4K))
      break;

    /* Maybe is the sector already formated ? Try with our key first to run faster */
    rc =
      SPROX_MifStUpdateAccessBlock(NULL, sector, MIFARE_KEY_B, MIFARE_KEY_A,
                                   MIFARE_KEY_B, ACC_BLOCK_VALUE,
                                   ACC_BLOCK_VALUE, ACC_BLOCK_VALUE,
                                   ACC_AUTH_NORMAL);
    if (rc != MI_OK)
      /* Let's try every key from reader's key set. Hopefully the card is in transport condition and the transport key is in reader's EEPROM... */
      rc =
        SPROX_MifStUpdateAccessBlock(NULL, sector, NULL, MIFARE_KEY_A,
                                     MIFARE_KEY_B, ACC_BLOCK_VALUE,
                                     ACC_BLOCK_VALUE, ACC_BLOCK_VALUE,
                                     ACC_AUTH_NORMAL);

    if (rc != MI_OK)
    {
      WriteLn("Failed to format sector " + IntToStr(sector));
      WriteRc(rc);
      return false;
    }
  }

  /* Put our header in sector 0 */
  memcpy(&buffer[16], MIFARE_HEADER, 16);
  memset(&buffer[32], 0, 16);
  rc = SPROX_MifStWriteSector(NULL, 0, buffer, MIFARE_KEY_B);
  if (rc != MI_OK)
  {
    WriteLn("Failed to write sector 0");
    WriteRc(rc);
    return false;
  }

  return true;
}

bool __fastcall TWorker::EraseCardMifare()
{
  unsigned char buffer[240];
  unsigned char sector;
  short rc;

  memset(buffer, 0, sizeof(buffer));

  WriteLn("Erasing Mifare card");
  WriteLn("... be patient, this may take a long time ...");

  /* Put default keys in each sector */
  for (sector = 0; sector < 40; sector++)
  {
    if ((sector >= 16) && (fMain->card_data.is_type != CARD_MIFARE_4K))
      break;

    /* Logically the sector is formated with our key */
    rc =
      SPROX_MifStUpdateAccessBlock(NULL, sector, MIFARE_KEY_B,
                                   MIFARE_TRANSPORT_KEY, MIFARE_TRANSPORT_KEY,
                                   ACC_BLOCK_VALUE, ACC_BLOCK_VALUE,
                                   ACC_BLOCK_VALUE, ACC_AUTH_NORMAL);
    if (rc != MI_OK)
      /* Failed ? Let's try every key from reader's key set. */
      rc =
        SPROX_MifStUpdateAccessBlock(NULL, sector, NULL, MIFARE_TRANSPORT_KEY,
                                     MIFARE_TRANSPORT_KEY, ACC_BLOCK_VALUE,
                                     ACC_BLOCK_VALUE, ACC_BLOCK_VALUE,
                                     ACC_AUTH_NORMAL);

    if (rc != MI_OK)
    {
      WriteLn("Failed to format sector " + IntToStr(sector));
      WriteRc(rc);
      return false;
    }
  }

  /* Put blank data in each sector */
  for (sector = 0; sector < 40; sector++)
  {
    if ((sector >= 16) && (fMain->card_data.is_type != CARD_MIFARE_4K))
      break;

    /* The card is in transport condition, we hope the transport key is in reader's EEPROM... */
    rc = SPROX_MifStWriteSector(NULL, sector, buffer, NULL);
    if (rc != MI_OK)
    {
      WriteLn("Failed to write sector " + IntToStr(sector));
      WriteRc(rc);
      return false;
    }
  }

  return true;
}

/*
 ****************************************************************************
 *
 *  MIFARE ULTRALIGHT
 *
 ****************************************************************************
 */

bool __fastcall TWorker::ReadCardMifareUL()
{
  WriteLn("Support of Mifare UltraLight is not yet implemented");
  return false;
}

bool __fastcall TWorker::WriteCardMifareUL()
{
  WriteLn("Support of Mifare UltraLight is not yet implemented");
  return false;
}

bool __fastcall TWorker::EraseCardMifareUL()
{
  WriteLn("Support of Mifare UltraLight is not yet implemented");
  return false;
}

/*
 ****************************************************************************
 *
 *  JAYCOS
 *
 ****************************************************************************
 */

#define JAYCOS_APPLICATION_ID   0x0123
#define JAYCOS_IDENT_FILE_ID    0x0124
#define JAYCOS_FINGER_1_FILE_ID 0x0125
#define JAYCOS_FINGER_2_FILE_ID 0x0126

bool __fastcall TWorker::ReadCardJaycos()
{
  unsigned char buffer[4096];
  unsigned long length;
  int i;
  short rc;

  WriteLn("Reading Jaycos card");

  /* Select root application in the Jaycos card */
  WriteLn("Selecting master file");
  rc = JaycosSelectMF();
  if (rc != MI_OK)
  {
    WriteLn("Card fatal error");
    WriteRc(rc);
    return false;
  }

  /* Select our application in the Jaycos card */
  WriteLn("Entering application directory");
  rc = JaycosSelectDF(JAYCOS_APPLICATION_ID);
  if (rc != MI_OK)
  {
    WriteLn("Card application not found");
    WriteRc(rc);
    return false;
  }

  /* Read the information file. This one is mandatory */
  WriteLn("Reading information file");
  rc = JaycosSelectEF(JAYCOS_IDENT_FILE_ID);
  if (rc != MI_OK)
  {
    WriteLn("File not found");
    WriteRc(rc);
    return false;
  }
  rc = JaycosReadBinary(0, IDENT_SIZE, buffer);
  if (rc != MI_OK)
  {
    WriteLn("Read from card failed");
    WriteRc(rc);
    return false;
  }

  /* Translate information file into interesting content... */
  DecodeUserData(buffer);

  /* Read the fingerprint files (if some) */
  if (fMain->card_data.finger_size[0])
  {
    WriteLn("Reading first fingerprint file");
    rc = JaycosSelectEF(JAYCOS_FINGER_1_FILE_ID);
    if (rc != MI_OK)
    {
      WriteLn("File not found");
      WriteRc(rc);
      return false;
    }
    rc = JaycosReadBinary(0, FINGERS_SIZE / 2, buffer);
    if (rc == MI_OK)
    {
      memcpy(fMain->card_data.finger_data[0], buffer, FINGERS_SIZE / 2);
    } else
    {
      WriteLn("Read from card failed");
      fMain->card_data.finger_size[0] = 0;
      WriteRc(rc);
      return false;
    }
  }
  if (fMain->card_data.finger_size[1])
  {
    WriteLn("Reading second fingerprint file");
    rc = JaycosSelectEF(JAYCOS_FINGER_2_FILE_ID);
    if (rc != MI_OK)
    {
      WriteLn("File not found");
      WriteRc(rc);
      return false;
    }
    rc = JaycosReadBinary(0, FINGERS_SIZE / 2, buffer);
    if (rc == MI_OK)
    {
      memcpy(fMain->card_data.finger_data[1], buffer, FINGERS_SIZE / 2);
    } else
    {
      WriteLn("Read from card failed");
      fMain->card_data.finger_size[1] = 0;
      WriteRc(rc);
      return false;
    }
  }

  return true;
}

bool __fastcall TWorker::WriteCardJaycos()
{
  unsigned char buffer[4096];
  unsigned long length;
  short rc;

  WriteLn("Writing Jaycos card");

  /* Select root application in the Jaycos card */
  WriteLn("Selecting master file");
  rc = JaycosSelectMF();
  if (rc != MI_OK)
  {
    WriteLn("Card fatal error");
    WriteRc(rc);
    return false;
  }

  /* Select our application in the Jaycos card */
  WriteLn("Entering application directory");
  rc = JaycosSelectDF(JAYCOS_APPLICATION_ID);
  if (rc != MI_OK)
  {
    /* Application doesn't exist, let's create it ! */
    if (!FormatCardJaycos())
      return false;

    /* Assume we're still in application directory */
  }

  /* Write the fingerprint files */
  if (fMain->card_data.finger_size[0])
  {
    WriteLn("Writing first fingerprint...");
    rc = JaycosSelectEF(JAYCOS_FINGER_1_FILE_ID);
    if (rc != MI_OK)
    {
      WriteLn("File not found");
      WriteRc(rc);
      return false;
    }
    rc =
      JaycosUpdateBinary(0, FINGERS_SIZE / 2,
                         fMain->card_data.finger_data[0]);
    if (rc != MI_OK)
    {
      WriteLn("Write to card failed");
      WriteRc(rc);
      return false;
    }
  }
  if (fMain->card_data.finger_size[1])
  {
    WriteLn("Writing second fingerprint...");
    rc = JaycosSelectEF(JAYCOS_FINGER_2_FILE_ID);
    if (rc != MI_OK)
    {
      WriteLn("File not found");
      WriteRc(rc);
      return false;
    }
    rc =
      JaycosUpdateBinary(0, FINGERS_SIZE / 2,
                         fMain->card_data.finger_data[1]);
    if (rc != MI_OK)
    {
      WriteLn("Write to card failed");
      WriteRc(rc);
      return false;
    }
  }

  /* Build the information file content */
  EncodeUserData(buffer);

  /* Write the information file */
  WriteLn("Writing information file...");
  rc = JaycosSelectEF(JAYCOS_IDENT_FILE_ID);
  if (rc != MI_OK)
  {
    WriteLn("File not found");
    WriteRc(rc);
    return false;
  }
  rc = JaycosUpdateBinary(0, IDENT_SIZE, buffer);
  if (rc != MI_OK)
  {
    WriteLn("Write to card failed");
    WriteRc(rc);
    return false;
  }

  return true;
}

bool __fastcall TWorker::FormatCardJaycos()
{
  WriteLn("Formatting Jaycos cards is not yet implemented");
  return false;
}


bool __fastcall TWorker::EraseCardJaycos()
{
  WriteLn("Erasing Jaycos cards is not yet implemented");
  return false;
}























bool __fastcall TWorker::ReadCardTask()
{
  /* Clear current data */
  fMain->card_data.has_user = false;
  fMain->card_data.picture_size = 0;
  fMain->card_data.finger_size[0] = 0;
  fMain->card_data.finger_size[1] = 0;

  switch (fMain->card_data.is_type)
  {
    case CARD_MIFARE_1K:
      return ReadCardMifare1k();
    case CARD_MIFARE_4K:
      return ReadCardMifare4k();
    case CARD_MIFARE_UL:
      return ReadCardMifareUL();
    case CARD_DESFIRE:
      return ReadCardDesfire();
    case CARD_JAYCOS:
      return ReadCardJaycos();
  }

  return false;
}

bool __fastcall TWorker::WriteCardTask()
{
  switch (fMain->card_data.is_type)
  {
    case CARD_MIFARE_1K:
      return WriteCardMifare1k();
    case CARD_MIFARE_4K:
      return WriteCardMifare4k();
    case CARD_MIFARE_UL:
      return WriteCardMifareUL();
    case CARD_DESFIRE:
      return WriteCardDesfire();
    case CARD_JAYCOS:
      return WriteCardJaycos();
  }

  return false;
}

bool __fastcall TWorker::EraseCardTask()
{
  switch (fMain->card_data.is_type)
  {
    case CARD_MIFARE_1K:
    case CARD_MIFARE_4K:
      return EraseCardMifare();
    case CARD_MIFARE_UL:
      return EraseCardMifareUL();
    case CARD_DESFIRE:
      return EraseCardDesfire();
    case CARD_JAYCOS:
      return EraseCardJaycos();
  }

  return false;
}

void __fastcall TWorker::Execute()
{
  bool rc;

  for (;;)
  {
    /* We wait for a command from the main loop */
    WaitForSingleObject(hWakeupEvent, INFINITE);

    /* What shall we do now ? */
    switch (nextTaskCode)
    {
      case WORK_TERMINATE:
        Terminate();            /* Hard exit... */
        return;

      case WORK_OPEN_READER:
        rc = OpenReaderTask();
        break;
      case WORK_FIND_CARD:
        rc = FindCardTask();
        break;
      case WORK_READ_CARD:
        rc = ReadCardTask();
        break;
      case WORK_WRITE_CARD:
        rc = WriteCardTask();
        break;
      case WORK_ERASE_CARD:
        rc = EraseCardTask();
        break;

      case WORK_ENROLL:
        rc = EnrollTask();
        break;
      case WORK_VERIFY:
        rc = VerifyTask();
        break;

      case WORK_CANCEL:
        rc = false;
        break;                  /* Not yet implemented */
    }

    /* Tell main form we're done */
    doneTaskCode = nextTaskCode;
    nextTaskCode = WORK_NONE;
    TaskDone(rc);
  }
}

void __fastcall TWorker::syncWriteLn()
{
  fProgress->btnClose->Enabled = false;
  fProgress->mVerbose->Lines->Add(syncLine);
  if (!fProgress->Visible)
    fProgress->Show();
}

void __fastcall TWorker::WriteLn(AnsiString S)
{
  syncLine = S;
  Synchronize(syncWriteLn);
}

void __fastcall TWorker::WriteRc(short rc)
{
  if (rc < 0)
  {
    WriteLn("Error code " + IntToStr(rc));
  } else if (rc > 0)
  {
    WriteLn("Error code 0x" + IntToHex(rc, 4));
  }
}

void __fastcall TWorker::syncTaskSuccess()
{
  /* Back to default cursor */
  Screen->Cursor = crDefault;
  fProgress->Repaint();
  fMain->Repaint();
  Application->ProcessMessages();

  /* Call fMain callback */
  fMain->OnWorkerDone(doneTaskCode, true);
}

void __fastcall TWorker::syncTaskFailure()
{
  /* Back to default cursor */
  Screen->Cursor = crDefault;
  fProgress->Repaint();
  fMain->Repaint();
  Application->ProcessMessages();

  /* Call fMain callback */
  fMain->OnWorkerDone(doneTaskCode, false);
}

void __fastcall TWorker::TaskDone(bool result)
{
  if (result)
    Synchronize(syncTaskSuccess);
  else
    Synchronize(syncTaskFailure);
}

void __fastcall TWorker::Cancel()
{
  Task(WORK_CANCEL);
}

void __fastcall TWorker::Task(work_task_e TaskCode)
{
  /* Hourglass cursor */
  Screen->Cursor = crHourGlass;
  Application->ProcessMessages();

  nextTaskCode = TaskCode;
  SetEvent(hWakeupEvent);
}

void __fastcall TWorker::DecodeUserData(unsigned char buffer[IDENT_SIZE])
{
  int i;

  fMain->card_data.has_user = true;

  /* Retrieve user data */
  memcpy(fMain->card_data.user_id, buffer, 8);

  TDate d;

  if (TryEncodeDate(1900 + buffer[13], buffer[14], buffer[15], d))
    fMain->card_data.birth_date = d.DateString();
  else
    fMain->card_data.birth_date = "";

  fMain->card_data.last_name = "";
  for (i = 0; i < 32; i++)
    fMain->card_data.last_name += (char) buffer[16 + i];
  fMain->card_data.first_name = "";
  for (i = 0; i < 32; i++)
    fMain->card_data.first_name += (char) buffer[48 + i];

  /* Get size of picture file */
  fMain->card_data.picture_size = buffer[8] * 0x0100 + buffer[9];

  /* Get type and size of fingerprint */
  fMain->card_data.finger_size[0] = buffer[10];
  fMain->card_data.finger_size[1] = buffer[11];
  fMain->card_data.finger_type = buffer[12];
}

void __fastcall TWorker::EncodeUserData(unsigned char buffer[IDENT_SIZE])
{
  memset(buffer, 0, IDENT_SIZE);

  memcpy(&buffer[0], fMain->card_data.user_id, 8);

  /* Size of picture */
  buffer[8] = (unsigned char) (fMain->card_data.picture_size / 0x0100);
  buffer[9] = (unsigned char) (fMain->card_data.picture_size % 0x0100);

  /* Information on fingerprint */
  buffer[10] = fMain->card_data.finger_size[0];
  buffer[11] = fMain->card_data.finger_size[1];
  buffer[12] = fMain->card_data.finger_type;

  /* Birth date */
  try
  {
    TDate d;
    unsigned short dy, dm, dd;

    d = StrToDate(fMain->card_data.birth_date);
    DecodeDate(d, dy, dm, dd);
    if ((dy >= 1900) && (dy <= 2050))
      buffer[13] = dy - 1900;
    if ((dm >= 1) && (dm <= 12))
      buffer[14] = dm;
    if ((dd >= 1) && (dd <= 31))
      buffer[15] = dd;
  }
  catch(...)
  {

  }

  /* Name */
  strncpy(&buffer[16], fMain->card_data.last_name.c_str(), 32);
  strncpy(&buffer[48], fMain->card_data.first_name.c_str(), 32);
}
